@@ -8,7 +8,7 @@ $ cd .../path/to/home-assistant/
 $ pip3 install --upgrade .../path/to/pybinarymoip
 
 Then the custom_component/pybinarymoip/ and its require line will work.
-(Or just make the custom_component's manifest.json requirements field 
+(Or just make the custom_component's manifest.json requirements field
 mention this dependency.)
 
 """
@@ -19,6 +19,7 @@ __copyright__ = "Copyright 2019, Greg J. Badros"
 import logging
 import socket
 import select
+import threading
 
 # urllib.disable_warnings(urllib3.exceptions.InsecureRequestWarning)
 _LOGGER = logging.getLogger(__name__)
@@ -44,7 +45,8 @@ class MoIP(object):
         self._transmitters = []
         self._devices = []
         self._sock = None
-        self._timeout = 3
+        self._timeout = 8
+        self._socket_lock = threading.Lock()
 
     def connect(self):
         """Begin connection to the device and get its status."""
@@ -99,27 +101,39 @@ class MoIP(object):
         except Exception as err:
             _LOGGER.error("Failed to initialize connection: %s", err)
 
+    # TODO: skip updating again if last update was < ~2 sec ago
     def _update_inputs(self):
-        self._send("?Receivers\n")
-        answer = self._read_after_equals()
-        inputs = answer.split(",")
-        for i in inputs:
-            (tx, rx) = i.split(":",1)
-            tx_obj = self._transmitters[int(tx)-1]
-            self._receivers[int(rx)-1]._set_input(tx_obj)
+        with self._socket_lock:
+            extra_text = self._read_full()
+            if extra_text:
+                _LOGGER.warning("Ignoring extra text: %r", extra_text)
+            self._send("?Receivers\n")
+            answer = self._read_after_equals()
+            if answer is None:
+                _LOGGER.error("Got bad response from ?Receivers "
+                              "-- cannot _update_inputs")
+                return
+            inputs = answer.split(",")
+            for i in inputs:
+                (tx, rx) = i.split(":", 1)
+                tx_obj = self._transmitters[int(tx)-1]
+                self._receivers[int(rx)-1]._set_input(tx_obj)
 
     def _send(self, str):
         self._last_send = str
         self._sock.send(str.encode())
 
-    def _send_check(self, str):
+    def _send_check(self, str, timeout):
         """Appends newline to str before sending, confirms OK response."""
-        self._send(str + "\n")
-        response = self._read()
-        _LOGGER.debug("sent '%s', got response = %s", str, response)
-        if response != "OK":
-            _LOGGER.error("Sent '%s' and got error response: %s",
-                          str, response)
+        if timeout is None:
+            timeout = self._timeout
+        with self._socket_lock:
+            self._send(str + "\n")
+            response = self._read(timeout)
+            _LOGGER.debug("sent '%s', got response = %s", str, response)
+            if response and not response.startswith("OK"):
+                _LOGGER.error("Sent '%s' and got error response: %s",
+                              str, response)
 
     def _read_raw(self, timeout=None):
         if timeout is None:
@@ -131,8 +145,10 @@ class MoIP(object):
         answer = self._sock.recv(16384).decode()
         return answer
 
-    def _read(self):
-        answer = self._read_raw()
+    def _read(self, timeout=None):
+        if timeout is None:
+            timeout = self._timeout
+        answer = self._read_raw(timeout)
         if not answer:
             _LOGGER.warning(
                 "Timeout (%s second(s)) waiting for a response after "
@@ -156,7 +172,11 @@ class MoIP(object):
 
     def _read_after_equals(self):
         answer = self._read()
-        return answer.split("=", 1)[1]
+        split_answer = answer.split("=", 1)
+        if len(split_answer) > 1:
+            return split_answer[1]
+        else:
+            return None
 
     @property
     def receivers(self):
@@ -204,14 +224,15 @@ class MoIP_Receiver(object):
     def _set_input(self, input):
         self._input = input
 
-    def _send_check(self, str):
-        return self._mc._send_check(str)
+    def _send_check(self, str, timeout=5):
+        return self._mc._send_check(str, timeout)
 
     def switch_to_tx(self, tx):
         if not isinstance(tx, int):
             tx = tx.num
+        self._input = tx
         self._send_check("!Switch=%s,%s" %
-                         (tx, self._num))
+                         (tx, self._num), 12)  # long timeout
         self._mc._update_inputs()
 
     def set_resolution(self, resolution):
@@ -242,7 +263,7 @@ class MoIP_Receiver(object):
         return str({'name': self._name,
                     'num': self._num,
                     'input': self._input,
-                    'input_num': self._input and self._input.num,
+#                    'input_num': self._input and self._input.num,
                     'mc': self._mc})
 
 
